#! /bin/sh
# \
exec tclsh${TCL} "$0" ${1+"$@"}

#
# AitCS Example
#
# Using an AitCS for "global" configuration data
#
# Stuart Cassoff
# Fall 2020
#


package require aitcs

namespace eval ::aitcs { namespace eval examples { namespace eval cfg {

proc printCfg {{msg ""}} {
	puts [format %-8s $msg][join [lmap field [cfg fields] { format {%s:%2d} $field [cfg $field] }] "  "]
}


proc adjustCfg {} {
	# Verify and adjust as necessary
	if {[expr {
		   [cfg nThings] < [cfg minThings]
		|| [cfg nThings] > [cfg maxThings]
	}]} {
		cfg nThings [cfg defaultThings]
	}
}


proc modifyCfg {} {
	# Incrementing by the max allowed
	# should put this out-of-bounds

	cfg+= nThings [cfg maxThings]
}


proc loadCfg {} {
	# Pretend to load cfg data

	cfg nThings 5
}


proc initCfg {} {
	# Create and set initial values
	#
	#                         
	# Create command     Name  EndOpts   Initial values
	# v                  v     v         v
	::aitcs::aitcs+ new  cfg   --        \
		fields [list nThings minThings maxThings defaultThings] \
		minThings 4 maxThings 10 defaultThings 6

	cfg nThings [cfg defaultThings]
}


proc go {} {
	initCfg   ; printCfg New
	loadCfg   ; printCfg Load
	modifyCfg ; printCfg Modify
	adjustCfg ; printCfg Adjust
}


}}}; # ::aitcs::examples::cfg


if {[info exists ::argv0] && [info script] eq $::argv0} {
	::aitcs::examples::cfg::go {*}$::argv
}


# EOF
