#
# AitCS - Array in the Command's Shadow
#
# Stuart Cassoff
#
#
# Spring 2018
# Version 1.1
#
# Winter 2020
# Version 1.2
#
# Spring 2020
# Version 1.3, 1.4, 1.5, 1.6, 1.7, 1.8
#
# Fall 2020
# Version 1.9
# Version 2.0b1
#


if {$::tcl_platform(engine) eq "Tcl"} { package require Tcl 8.6 }

namespace eval aitcs {

proc pkgconfig {subcommand args} {
	if {$subcommand ni [list get list]} {
		return -code error "unknown subcommand \"$subcommand\": must be get, or list"
	}
	set pkgconfig [dict create {*}{
		name	aitcs
		title	AitCS
		descr	{Array in the Command's Shadow}
		author	{Stuart Cassoff}
		when	{Fall 2020}
		version	2.0b1
	}]
	dict set pkgconfig license,type ISC
	dict set pkgconfig license,text [regsub -all {^\s+|\s+$|\t} {
		Copyright (c) 2018,2020 Stuart Cassoff <stwo@users.sourceforge.net>

		Permission to use, copy, modify, and distribute this software for any
		purpose with or without fee is hereby granted, provided that the above
		copyright notice and this permission notice appear in all copies.

		THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
		WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
		MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
		ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
		WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
		ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
		OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
	} {}]
	return -code {*}[expr {
	    $subcommand eq "list"
		? [llength $args] == 0
		    ? [list ok [dict keys $pkgconfig]]
		    : [list error "wrong # args: should be \"[namespace current]::pkgconfig list\""]
		: $subcommand eq "get"
		    ? [llength $args] == 1
			? [dict exists $pkgconfig [set key [lindex $args 0]]]
			    ? [list ok [dict get $pkgconfig $key]]
			    : [list error "key \"$key\" not known"]
			: [list error "wrong # args: should be \"[namespace current]::pkgconfig get key\""]
		    : [list uh oh]
	}]
}


## Internal

if {$::tcl_platform(engine) eq "Jim"} {
	proc Alias {this that args} {        alias    $this    $that {*}$args }
} else {
	proc Alias {this that args} { interp alias {} $this {} $that {*}$args }
}

proc Randstr {{length {6}}} {
	set str ""
	for {set i 0} {$i < $length} {incr i} {
		append str [format %c [expr {97 + int(rand() * 26)}]]
	}
	return $str
}

proc Randvar {name variable_namespace} {
	return [string trimright $variable_namespace :]::__${name}__[Randstr]_
}

proc MakeCommands {type name varname cns} {
	set jim [expr {$::tcl_platform(engine) eq "Jim"}]

	set cns [string trimright $cns :]
	set smae [expr {[namespace qualifiers $varname] eq $cns}]
	set cns $cns:

	set ops   [list ""   ?       -      ^    &    ->    ]
	set procs [list Vset Vexists Vunset Vgwd Vref Vaitcs]

	if {$type eq "eagle"} {
		lappend ops   =      +        +=
		lappend procs Vsetto Vlappend Vincr
	}

	if {$jim} {
		set hows [lrepeat [llength $ops] full]
	} else {
		if {$smae} {
			set hows [lrepeat [llength $ops] var]
		} else {
			set hows [lrepeat [llength $ops] full]
			set pl [list Vexists Vgwd]
			if {$type eq "eagle"} {
				lappend pl Vsetto
			}
			foreach p $pl {
				lset hows [lsearch $procs $p] partial
			}
		}
		lset hows [lsearch $procs Vref] full
	}

	foreach op $ops proc $procs how $hows {
		set pargs [info args $proc]
		if {!$jim} {
			set pargs [lmap arg $pargs {
				expr {[info default $proc $arg def] ? [list $arg $def] : $arg}
			}]
		}
		set body [info body $proc]
		if {[lindex $pargs 0] eq "varname"} {
			set pargs [lrange $pargs 1 end]
			# For itiming tests, consult Makefile.aitcs.in when modifying this line: set how ?
			set body [switch -exact -- $how {
				var { # Change "upvar ..." to "variable ..." and hardcode [namespace tail $varname] everywhere
					set vn [namespace tail $varname]
					string map [list {upvar 1 $varname _v_} "variable $vn" _v_ $vn {$varname} $vn {${varname}} $vn] $body
				}
				full { # Remove "upvar ..." and hardcode $varname everywhere
					string map [list {upvar 1 $varname _v_} {} _v_ $varname {$varname} $varname {${varname}} $varname] $body
				}
				partial { # Hardcode $varname where used
					string map [list {$varname} $varname {${varname}} $varname] $body
				}
				setvar { # Add "set varname $varname" to top of proc
					format {set varname {%s}%s%s} $varname \n $body
				}
				alias { # Use aliases
					if {$proc eq "Vaitcs"} {
						if {[llength [info commands Vaitcs_alias]] == 0} {
							proc Vaitcs_alias {varname elem args} [string map [list \
								{tailcall $_v_($elem)$op {*}$args} \
								{uplevel 1 $_v_($elem)$op $args}] $body]
						}
						#set proc [string cat $proc _alias]
						append proc _alias
					}
					Alias $cns:$name$op [namespace current]::$proc $varname
					continue
				}
			}]
		}

		# Debug
		#puts "$name$op {$pargs} {$body\n}"

		proc $cns:$name$op $pargs $body

		# Debug
		#puts $name$op\n[::tcl::unsupported::disassemble proc $cns:$name$op]
	}

	# Debug
	#exit

	Alias $cns:$name! upvar #0 $varname $name

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name&&    return -level 0 $cns:$name
	proc    $cns:$name&& {} [list return $cns:$name]

	set --- [join [list {} \
		"unset $varname" \
		{*}[lmap op [list {*}$ops ! && ---] {
			format {rename %s%-3s {}} $cns:$name $op
		}] \
	] \n\t]\n

	# For itiming tests, consult Makefile.aitcs.in when modifying these 3 lines
	# Alias $cns:$name--- eval ${---}
	proc    $cns:$name--- {}   ${---}
}

proc New {type args} {
	set name ""
	set cns ""
	set vns ""
	set argies [list]
	set nameset no
	set state opt
	foreach arg $args {
		switch -exact -- $state {
		opt {
			switch -exact -- $arg {
			-cns - -command_namespace - -vns - -variable_namespace { set state $arg }
			-- { set state -- }
			default { if {$nameset} { lappend argies $arg } else { set name $arg; set nameset yes } }
			}; # hctiws
		}
		-cns - -command_namespace  { set state opt; set cns $arg }
		-vns - -variable_namespace { set state opt; set vns $arg }
		-- { lappend argies $arg }
		default { error "Huh? <[expr {int(rand()*100000)}]>" }
		}; # hctiws
	}
	if {$state ni [list opt --]} { return -code error "no argument given for option \"$state\"" }
	if {[expr {[llength $argies] % 2}] != 0} { return -code error "initial values must be \"key value ...\"" }
	if {$vns eq ""} { set vns [uplevel 1 {namespace current}] }
	if {$cns eq ""} { set cns $vns }
	set vns [string trimright $vns :]
	set cns [string trimright $cns :]
	if {$name eq ""} { for {*}{
		{ set name [Randstr]; set i 0 }
		{ $i < 1000 && [llength [info commands ${cns}::$name]] > 0 }
		{ set name [Randstr]; incr i }
		{}
	}}
	if {[llength [info commands ${cns}::$name]] > 0} { return -code error "${cns}::$name already exists" }
	set varname [Randvar $name $vns]
	MakeCommands $type $name $varname $cns
	array set $varname $argies
	return ${cns}::$name
}


## Operators

# AitCS / AitCS+

#
proc Vset {varname elem args} {
	upvar 1 $varname _v_
	if {[llength $args] == 0} {
		set _v_($elem)
	} elseif {[llength $args] == 1} {
		set _v_($elem) [lindex $args 0]
	} else {
		array set _v_ [list $elem {*}$args]
		set _v_($elem)
	}
}

# -
proc Vunset {varname elem args} {
	upvar 1 $varname _v_
	foreach e $args { array unset _v_ $e }
	array unset _v_ $elem
}

# &
proc Vref {varname {elem ""}} {
	if {$elem eq ""} { return $varname }
	return ${varname}($elem)
}

# ?
proc Vexists {varname elem {myVar ""}} {
	upvar 1 $varname _v_
	if {![info exists _v_($elem)]} { return 0 }
	if {$myVar eq {}} { return 1 }
	upvar 1 $myVar var
	set var $_v_($elem)
	return 1
}

# ^
proc Vgwd {varname elem {default ""} {myVar ""}} {
	upvar 1 $varname _v_
	if {$myVar ne ""} { upvar 1 $myVar var }
	set var [expr {[info exists _v_($elem)] ? $_v_($elem) : $default}]
}

# ->
proc Vaitcs {varname elem args} {
	upvar 1 $varname _v_
	# Avoid mismatches by checking double-char operators before single-char operators
	if {[set op [string range $elem end-1 end]] in [list += && ->]} {
		set elem [string range $elem 0 end-2]
	} elseif {[set op [string index $elem end]] in [list ? - ^ & = + ! :]} {
		set elem [string range $elem 0 end-1]
	} elseif {[set op [string range $elem end-2 end]] eq "---"} {
		set elem [string range $elem 0 end-3]
	} else {
		set op ""
	}
	tailcall $_v_($elem)$op {*}$args
}

# AitCS+

# =
proc Vsetto {varname elem elemOrVal args} {
	upvar 1 $varname _v_
	set a [list $elem $elemOrVal {*}$args]
	set val [lindex $a end]
	foreach e [lrange $a 0 end-1] { set _v_($e) $val }
	return $val
}

# +
proc Vlappend {varname elem args} {
	upvar 1 $varname _v_
	lappend _v_($elem) {*}$args
}

# +=
proc Vincr {varname elem {incr {1}} args} {
	upvar 1 $varname _v_
	foreach i $args { incr _v_($elem) $i }
	incr _v_($elem) $incr
}


## Setup

if {$::tcl_platform(engine) eq "Jim"} {
	foreach plus [list "" +] type [list snake eagle] {
		proc aitcs$plus {cmd args} [string map [list @plus@ $plus @type@ $type] {
			if {$cmd eq "operators"} {
				tailcall ::aitcs::$cmd@plus@ {*}$args
			} elseif {$cmd eq "new"} {
				tailcall ::aitcs::New @type@ {*}$args
			} else {
				return -code error "unknown subcommand \"$cmd\": must be new, or operators"
			}
		}]
	}
} else {
	foreach plus [list "" +] type [list snake eagle] {
		namespace ensemble create -prefixes no \
			-command aitcs$plus \
			-map [dict create new [list New $type] operators operators$plus]
	}
}


# These things
proc operators  {} { return [list "" ? - ^ & -> ! && ---] }
proc operators+ {} { return [list {*}[operators] = + +=] }


namespace export aitcs aitcs+

package provide aitcs 2.0b1


}; # namespace eval aitcs


# EOF
